"""Cli command to make a reference"""

import logging

import click

from fluffy.config import check_configs
from fluffy.workflows.wisecondor import make_reference
from fluffy.status import print_status

LOG = logging.getLogger(__name__)


@click.command()
@click.option("--dry-run", is_flag=True, help="Do not create any files")
@click.pass_context
def reference(ctx, dry_run):
    """Create a reference for WisecondorX"""
    LOG.info("Running fluffy reference")
    configs = ctx.obj["configs"]
    try:
        check_configs(configs, mkref=True)
    except FileNotFoundError as err:
        raise click.Abort

    configs = ctx.obj["configs"]
    slurm_api = ctx.obj["slurm_api"]

    print_status(
        output_dir=configs["out"],
    )

    jobid = make_reference(
        samples=ctx.obj["samples"],
        configs=configs,
        slurm_api=slurm_api,
        dry_run=dry_run,
    )
    LOG.info("Running make reference on slurm with jobid %s", jobid)



"""Code to print analysis status json file"""

import logging
import json

from pathlib import Path
from typing import Iterator
import sys

from datetime import datetime
from fluffy.version import __version__

LOG = logging.getLogger(__name__)

def print_status(output_dir: Path) -> None:
    """Create a status json file"""

    analysis_time=datetime.now()
    time_string=f"{analysis_time.year}-{analysis_time.month}-{analysis_time.day}T{analysis_time.hour}:{analysis_time.minute}:{analysis_time.second}"

    analysis_status={"analysis_run_status":"running","analysis_command":" ".join(sys.argv),"analysis_time":time_string,"fluffy_version":__version__}
    f=open(f"{str(output_dir)}/analysis_status.json","w")
    f.write(json.dumps(analysis_status))
    f.close()
    return(analysis_status)





"""Workflows for the preface program"""

import logging
from pathlib import Path
from typing import Iterator

from slurmpy import Slurm
from fluffy.singularity_cmd import singularity_base
from fluffy.commands.preface import (get_preface_model_cmd,
                                     get_preface_predict_cmd)
from fluffy.slurm_api import SlurmAPI

LOG = logging.getLogger(__name__)

def preface_predict_workflow(
    configs: dict,
    sample_id: str,
    afterok: int,
    slurm_api: SlurmAPI,
    dry_run: bool = False,
):
    """Run the preface predict workflow"""
    LOG.info("Running the preface predict workflow")
    out_dir = configs["out"]

    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])

    preface_predict_cmd = get_preface_predict_cmd(
        singularity=singularity,
        out_dir=out_dir,
        model_dir=configs["preface"]["model_dir"],
        sample_id=sample_id,
    )

    jobid = slurm_api.run_job(
        name=f"preface_predict-{sample_id}",
        command=preface_predict_cmd,
        afterok=[afterok],
        dry_run=dry_run,
    )

    return jobid



"""Workflow for analysing all samples in a batch"""
import copy

from typing import Iterator

from fluffy.slurm_api import SlurmAPI
from fluffy.workflows.align import align_individual
from fluffy.workflows.amycne import estimate_ffy
from fluffy.workflows.cleanup import cleanup_workflow
from fluffy.workflows.picard import picard_qc_workflow
from fluffy.workflows.preface import preface_predict_workflow
from fluffy.workflows.summarize import summarize_workflow
from fluffy.workflows.wisecondor import wisecondor_xtest_workflow
from fluffy.workflows.status_update import pipe_complete
from fluffy.workflows.status_update import pipe_fail


def analyse_workflow(
    samples: Iterator[dict],
    configs: dict,
    slurm_api: SlurmAPI,
    skip_preface: bool = False,
    dry_run: bool = False,
) -> int:
    """Run the wisecondor analysis"""
    jobids = []
    for sample in samples:
        sample_jobids = []
        sample_id = sample["sample_id"]
        sample_outdir = configs["out"] / sample_id
        # This will fail if dir already exists
        sample_outdir.mkdir(parents=True)

        slurm_api.slurm_settings["ntasks"]=configs["align"]["ntasks"]
        slurm_api.slurm_settings["mem"]=configs["align"]["mem"]

        align_jobid = align_individual(
            configs=configs, sample=sample, slurm_api=slurm_api, dry_run=dry_run,
        )

        slurm_api.slurm_settings["ntasks"]=configs["slurm"]["ntasks"]
        slurm_api.slurm_settings["mem"]=configs["slurm"]["mem"]

        ffy_jobid = estimate_ffy(
            configs=configs,
            sample_id=sample_id,
            afterok=align_jobid,
            slurm_api=slurm_api,
            dry_run=dry_run,
        )
        jobids.append(ffy_jobid)
        sample_jobids.append(ffy_jobid)

        picard_jobid = picard_qc_workflow(
            configs=configs,
            sample_id=sample_id,
            afterok=align_jobid,
            slurm_api=slurm_api,
            dry_run=dry_run,
        )
        jobids.append(picard_jobid)
        sample_jobids.append(picard_jobid)

        wcx_test_jobid = wisecondor_xtest_workflow(
            configs=configs,
            sample_id=sample_id,
            afterok=align_jobid,
            slurm_api=slurm_api,
            dry_run=dry_run,
        )

        jobids.append(wcx_test_jobid)
        sample_jobids.append(wcx_test_jobid)

        if not skip_preface:
            preface_predict_jobid = preface_predict_workflow(
                configs=configs,
                sample_id=sample_id,
                afterok=wcx_test_jobid,
                slurm_api=slurm_api,
                dry_run=dry_run,
            )
            jobids.append(preface_predict_jobid)
            sample_jobids.append(preface_predict_jobid)

        cleanup_jobid=cleanup_workflow(
            configs=configs,
            sample_outdir=sample_outdir,
            sample_id=sample_id,
            afterok=sample_jobids,
            slurm_api=slurm_api,
            dry_run=dry_run
        )
    jobids.append(cleanup_jobid)
 
    summarize_jobid = summarize_workflow(
        configs=configs, afterok=jobids, slurm_api=slurm_api, dry_run=dry_run
    )
    slurm_api.print_submitted_jobs()

    slurm_api.slurm_settings["time"]="1:00:00"
    pipe_complete(
        configs=configs, afterok=summarize_jobid, slurm_api=slurm_api, dry_run=dry_run
    )

    pipe_fail(
        configs=configs, slurm_api=slurm_api, dry_run=dry_run, afternotok=summarize_jobid
    )

    return summarize_jobid



"""An API to slurm"""
import copy
import yaml
import time as pytime
from datetime import datetime

import logging
from pathlib import Path

from slurmpy import Slurm

LOG = logging.getLogger(__name__)

class SlurmAPI:
    """An API to SLURM"""

    def __init__(self,slurm_settings: dict, out_dir: Path):
        super(SlurmAPI, self).__init__()
        LOG.info("Initializing a slurm API")
        account=None
        if "account" in slurm_settings:
            account=slurm_settings["account"]
        if not isinstance(account, str):
            raise SyntaxError("Invalid account {}".format(account))
        self.account = account

        time=None
        if "time" in slurm_settings:
            time=slurm_settings["time"]

        if not isinstance(time, str):
            raise SyntaxError("Invalid time {}".format(time))
        if not len(time.split(":")) == 3:
            LOG.warning("Specify time on format hh:mm:ss")
            raise SyntaxError("Invalid time format {}".format(time))
        self.time = time
        if not isinstance(out_dir, Path):
            raise SyntaxError("Invalid out dir format {}".format(out_dir))

        self.log_dir = out_dir / "logs"
        self.sacct_dir = out_dir / "sacct"
        self.scripts_dir = out_dir / "scripts"
        self.out_dir=out_dir
        self.slurm_settings=copy.copy(slurm_settings)
        self.job = None
        self.jobids=[]

        current_time=datetime.now()
        self.analysis_time=f"{current_time.year}-{current_time.month}-{current_time.day}T{current_time.hour}:{current_time.minute}:{current_time.second}"

    def create_job(self, name: str, afterok: list = None, afternotok: list = None) -> Slurm:
        """Create a job for submitting to SLURM"""
        LOG.info("Create a slurm job with name %s", name)

        self.slurm_settings["dependency"]=[]
        if afterok:
            self.slurm_settings["dependency"].append( "afterok:{}".format( ":".join( str(dependency) for dependency in afterok)  ) )
        if afternotok:
            self.slurm_settings["dependency"].append( "afternotok:{}".format( ":".join( str(dependency) for dependency in afternotok)  ) )

        if self.slurm_settings["dependency"]:
            self.slurm_settings["dependency"]=",".join(self.slurm_settings["dependency"])
        else:
            del self.slurm_settings["dependency"]

        job = Slurm(
            name,
            self.slurm_settings,
            log_dir=str(self.log_dir),
            scripts_dir=str(self.scripts_dir),
        )
        return job

    def run_job(
        self, name: str, command: str, afterok: list = None, afternotok: list = None, dry_run: bool = False
    ) -> int:
        """Create and submit a job to slurm"""
        job = self.create_job(name=name,afterok=afterok,afternotok=afternotok)
        LOG.info("Submitting commands %s", command)
        if afterok:
            LOG.info("Adding dependencies: %s", ",".join( str(dependency) for dependency in afterok))
        jobid = 1

        time_string=self.analysis_time
        on_finnish=""
        if len(self.jobids):
            #pytime.sleep(1)
            on_finnish="""
finnish(){{
sacct --format=jobid,jobname%50,account,partition,alloccpus,TotalCPU,elapsed,start,end,state,exitcode --jobs {} | perl -nae \'my @headers=(jobid,jobname,account,partition,alloccpus,TotalCPU,elapsed,start,end,state,exitcode); if($. == 1) {{ print q{{#}} . join(qq{{\\t}}, @headers), qq{{\\n}} }} if ($. >= 3 && $F[0] !~ /( .batch | .bat+ )\\b/xms) {{ print join(qq{{\\t}}, @F), qq{{\\n}} }}\' > {}/fluffy_{}.log.status
}}

failure(){{
sed -i \'s/ \"running\"/ \"fail\"/g\' {}/analysis_status.json
}}

trap finnish EXIT TERM INT
trap failure ERR TERM

""".format(",".join(self.jobids),self.sacct_dir,time_string,self.out_dir )

        if not dry_run:               
            jobid = job.run("\n".join([on_finnish, command]))
        LOG.info("Submitted job %s with job id: %s", name, jobid)

        self.jobids.append( str(jobid) )

        return jobid

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(account={self.account!r}, time={self.time!r}, "
            f"log_dir={self.log_dir!r}, scripts_dir={self.scripts_dir!r})"
        )


    def print_submitted_jobs(self,dry_run: bool = False):

        project_id=str(self.out_dir).strip("/").split("/")[-1]
        try:
            yaml_out=yaml.dump({project_id:[int(i) for i in self.jobids]})
        except:
            yaml_out=yaml.dump({project_id:self.jobids})
        if not dry_run:
            f=open("{}//submitted_jobs.yaml".format(self.sacct_dir),"w")
            f.write(yaml_out)
            f.close()



"""Make a cleaned up per sample zip file"""

from pathlib import Path

from fluffy.slurm_api import SlurmAPI
from fluffy.version import __version__
from fluffy.commands.multiqc import get_multiqc_cmd
from fluffy.singularity_cmd import singularity_base

def get_cleanup_cmd(out_dir: Path, sample_outdir: Path, sample_id: str) -> str:
    """Return a string with the command to clean up and compress a sample run folder"""
    sample_results_file = f"{str(out_dir)}/{sample_id}.fluffy-{__version__}"

    cleanup_cmd = (
        f"cp -r {sample_outdir} {sample_results_file}; "
        f"rm -f {sample_results_file}/*.sai {sample_results_file}/*.bam; "
        f"zip -r {sample_results_file}.zip {sample_results_file}; "
        f"rm -rf {sample_results_file}"
    )
    return cleanup_cmd


def cleanup_workflow(
    configs: dict,
    sample_outdir: Path,
    sample_id: str,
    afterok: list,
    slurm_api: SlurmAPI,
    dry_run: bool = False,
) -> int:
    """Run the workflow to compress  an analysis folder"""

    out_dir = configs["out"]
    cleanup_cmd = get_cleanup_cmd(
        out_dir=out_dir, sample_outdir=sample_outdir, sample_id=sample_id,
    )

    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])

    multiqc_cmd=get_multiqc_cmd(singularity=singularity,input_dir=sample_outdir,out_dir=sample_outdir)

    jobid = slurm_api.run_job(
        name=f"cleanup-{sample_id}",
        command="\n".join([multiqc_cmd,cleanup_cmd]),
        afterok=afterok,
        dry_run=dry_run,
    )

    return jobid
